require 'strscan'

Scanner = StringScanner.new($stdin.read)
Functions = Hash.new

def getToken
	Scanner.scan(/\s*(([\w<>]+|\[|\]|\(|\)|#|=|,))/)
	Scanner[1]
end

class Function

attr_reader :args, :name

def initialize
	@args = Hash.new
	@label = 0
end

def expr
	return nil unless tok = getToken
	case tok
		when '#'
			puts "\txor eax,eax"
			return true
		when 'left'
			return nil unless expr
			puts "\tcall left"
			return true
		when 'right'
			return nil unless expr
			puts "\tcall right"
			return true
		when '['
			return nil unless expr
			puts "\tpush eax"
			return nil unless getToken == ','
			return nil unless expr
			puts "\tpop ecx"
			puts "\tcall tuple"
			return nil unless getToken == ']'
			return true
		when 'if'
			return nil unless expr
			lelse = @label
			ldone = @label + 1
			@label += 2
			puts "\ttest eax,eax"
			puts "\tjz .l#{lelse}"
			return nil unless getToken == 'then'
			return nil unless expr
			puts "\tjmp .l#{ldone}"
			puts ".l#{lelse}:"
			return nil unless getToken == 'else'
			return nil unless expr
			puts ".l#{ldone}:"
			return true
		else
			if @args[tok]
				puts "\tmov eax,[ebp+8+4*#{@args.size-1-@args[tok]}]"
				return true
			elsif Functions[tok]
				f = Functions[tok]
				if f.args.size == 0
					puts "\tcall #{f.name}"
					return true
				end
				return nil unless getToken == '('
				f.args.size.times {
					return nil unless expr
					puts "\tmov ecx,eax"
					puts "\tcall link"
					puts "\tpush eax"
					tok = getToken
					return nil unless (tok == ',') || (tok == ')')
				}
				puts "\tcall #{f.name}"
				f.args.size.times {
					puts "\tpop ecx"
					puts "\tcall unlink"
				}
				return true
			else
				return nil
			end
	end		
end

def compile
	return nil unless fname = getToken
	Functions[fname] = self
	@name = "fun_#{fname}"
	return nil if getToken != '('
	loop do
		return nil unless argname = getToken
		break if argname == ')'
		argname = getToken if argname == ','
		@args[argname] = @args.size
	end
	return nil if getToken != '='
	puts "public #{@name}"
	puts "#{@name}:"
	puts "\tpush ebp"
	puts "\tmov ebp,esp"
	return nil unless expr
	puts "\tmov esp,ebp"
	puts "\tpop ebp"
	puts "\tret\n"
	return true
end

end

def compile
	while tok = getToken
		if tok == 'fun'
			f = Function.new
			return nil unless f.compile
		end
	end
	return true
end

puts "format ELF"
puts "section '.text' executable"
puts "extrn tuple"
puts "extrn link"
puts "extrn unlink"
puts "extrn left"
puts "extrn right"

if compile
	$stderr.puts "Done."
	exit(0)
else
	$stderr.puts "Error."
	$stderr.puts "#{Scanner.pre_match[-30..-1]} ***** " +
		"#{Scanner.post_match[0..30]}"
	exit(1)
end


